/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@boo.net 9/24/08
--------------------------------------------------------------------*/

#ifndef _POLY_SKEW_H_
#define _POLY_SKEW_H_

#include <gmp_xface.h>
#include "poly.h"

#ifdef __cplusplus
extern "C" {
#endif

/* external interface to skewed polynomial selector */

typedef void (*stage1_callback_t)(mpz_t a5, mpz_t p, mpz_t m, void *extra);

typedef struct {
	mpz_t gmp_N;
	mpz_t gmp_a5_begin;
	mpz_t gmp_a5_end;
	uint32 npr_in_p;
	double norm_max;
	uint32 p0_limit;
	stage1_callback_t callback;
	void *callback_data;
} poly_stage1_t;

void poly_stage1_init(poly_stage1_t *data, 
			stage1_callback_t callback,
			void *callback_data);
void poly_stage1_free(poly_stage1_t *data);
uint32 poly_stage1_run(poly_stage1_t *data);
uint32 poly_stage1_monic_run(poly_stage1_t *data);


typedef void (*stage2_callback_t)(void *extra, int deg,
				mpz_t * coeff1, mpz_t * coeff2,
				double skewness, double norm, 
				double root_score, double combined_score);

typedef struct {
	mpz_t gmp_N;
	double area;
	double bound0;
	double bound1;
	double min_e;
	double max_norm_1;
	double max_norm_2;
	uint32 p_bound;

	void *internal;

	stage2_callback_t callback;
	void *callback_data;
} poly_stage2_t;

void poly_stage2_init(poly_stage2_t *data,
		      stage2_callback_t callback,
		      void *callback_data);
void poly_stage2_free(poly_stage2_t *data);
void poly_stage2_run(poly_stage2_t *data, mpz_t a5, mpz_t p, mpz_t d);

#ifdef __cplusplus
}
#endif

#endif /* _POLY_SKEW_H_ */
