#include "stage1.h"

/*------------------------------------------------------------------------*/
static void
search_p(curr_poly_t *curr, rat_factor_t *rat_factors,
		knapsack_data_t *knap, bounds_t *bounds, 
		stage1_stat_t *stats, unsigned int p0_max)
{
	int i;
	unsigned int r0;
	unsigned int p0 = 1;

	for (i = 0; i < rat_factors->p0_list_len; i++) {
		rat_factors->last_p0 = p0;
		p0 = rat_factors->p0_list[i].p;
		r0 = rat_factors->p0_list[i].r;
		if (p0 > p0_max)
			break;

		stats->n_p++;
		mpz_mul_ui(curr->gmp_prod, curr->gmp_prod_base, p0);

		if (p0 == 1) {
			init_knapsack_exact(curr, knap, rat_factors,
					bounds->a3_max);
		}
		else {
			init_knapsack_exact_p0(curr, knap, rat_factors,
					p0, r0, bounds->a3_max);
		}
		knapsack_exact(curr, knap, bounds, stats);
	}
}

/*------------------------------------------------------------------------*/
static void
search_a5(curr_poly_t *curr, rat_factor_t *rat_factors,
		knapsack_data_t *knap,
		bounds_t *bounds, stage1_stat_t *stats)
{
	int i;
	double p_size, dp0_max, dbl_a5, dbl_N;
	unsigned int p0_max;
	unsigned int shift = 0;
	mpz_t gmp_tmp;
	uint32 npr_in_p, npr_total;
	uint32 *curr_p_idx;
	uint32 *a5_p_idx;
	rat_prime_t *primes;

	uint32 first_entry, list_size;

	dbl_N = mpz_get_d(curr->gmp_N);
	dbl_a5 = mpz_get_d(curr->gmp_a5);
	stage1_bounds_update(bounds, dbl_N, dbl_a5);
	mpz_init_set(gmp_tmp, curr->gmp_a5);

	while (1) {
		stats->n_a5++;
		if (++shift >= 1000) {
			/* update some variables, not very precise 
			   at the beginning of a5-range */
			mpz_add_ui(gmp_tmp, gmp_tmp, shift * MULTIPLIER);
			if (mpz_cmp(gmp_tmp, bounds->gmp_a5_end) > 0)
				break;
			dbl_a5 = mpz_get_d(gmp_tmp);
			stage1_bounds_update(bounds, dbl_N, dbl_a5);
			shift = 0;
		}

		mpz_add_ui(curr->gmp_a5, curr->gmp_a5, MULTIPLIER);

		if (!rat_factor_find(rat_factors, curr->gmp_a5,
					bounds->p_size_max)) {
			continue;
		}

		if (mpz_cmp(curr->gmp_a5, bounds->gmp_a5_end) > 0)
			break;

		stats->n_pr++;
		mpz_fdiv_q(curr->gmp_root, curr->gmp_N, curr->gmp_a5);
		mpz_root(curr->gmp_root, curr->gmp_root, 5);
		mpz_set(curr->gmp_m0, curr->gmp_root);

		dbl_a5 = mpz_get_d(curr->gmp_a5);
		stage1_bounds_update(bounds, dbl_N, dbl_a5);

		primes = rat_factors->primes;
		a5_p_idx = rat_factors->a5_p_idx;
		curr_p_idx = rat_factors->curr_p_idx;
		npr_in_p = rat_factors->npr_in_p;
		npr_total = rat_factors->npr_total;

		for (i = 0, p_size = 0.0; i < npr_in_p; i++)
			p_size += primes[a5_p_idx[i]].logprime;
		rat_factor_fill_aux(rat_factors, 
				curr->gmp_N, curr->gmp_a5,
				exp(bounds->p_size_max - p_size),
				bounds->p0_limit);

		first_entry = 0;
		list_size = npr_in_p;

		while (1) {
			for (i = 0; i < list_size; i++) {
				curr_p_idx[npr_in_p - list_size + i] =
						a5_p_idx[first_entry + i];
			}

			for (i = 0, p_size = 0.0; i < npr_in_p; i++)
				p_size += primes[curr_p_idx[i]].logprime;

			if (p_size > bounds->p_size_max) {
				if (++list_size > npr_in_p)
					break;
				first_entry = primes[curr_p_idx[npr_in_p - 
						list_size]].my_idx + 1;
				continue;
			}

			dp0_max = exp(bounds->p_size_max - p_size);
			if (dp0_max > (double) (bounds->p0_limit))
				p0_max = bounds->p0_limit;
			else
				p0_max = (unsigned int) (rint(dp0_max) - 1);
			if (p0_max < 1)
				p0_max = 1;

			mpz_set_ui(curr->gmp_prod_base, 1);
			for (i = 0; i < npr_in_p; i++) {
				mpz_mul_ui(curr->gmp_prod_base, 
					curr->gmp_prod_base, 
					primes[curr_p_idx[i]].prime);
			}

			search_p(curr, rat_factors, knap,
					bounds, stats, p0_max);

			if (first_entry < npr_total - list_size)
				list_size = 0;
			list_size++;

			if (primes[curr_p_idx[0]].my_idx == 
						npr_total - npr_in_p) {
				break;
			}

			first_entry = primes[curr_p_idx[npr_in_p - 
						list_size]].my_idx + 1;
		}
	}

	mpz_clear(gmp_tmp);
}

/*------------------------------------------------------------------------*/
void
poly_stage1_init(poly_stage1_t *data,
		 stage1_callback_t callback, void *callback_data)
{
	memset(data, 0, sizeof(poly_stage1_t));
	mpz_init_set_ui(data->gmp_N, 0);
	mpz_init_set_ui(data->gmp_a5_begin, 0);
	mpz_init_set_ui(data->gmp_a5_end, 0);
	data->callback = callback;
	data->callback_data = callback_data;
}

/*------------------------------------------------------------------------*/
void
poly_stage1_free(poly_stage1_t *data)
{
	mpz_clear(data->gmp_N);
	mpz_clear(data->gmp_a5_begin);
	mpz_clear(data->gmp_a5_end);
}

/*------------------------------------------------------------------------*/
uint32
poly_stage1_run(poly_stage1_t *data)
{
	uint32 success = 0;
	stage1_stat_t stats;
	bounds_t bounds;
	curr_poly_t currpoly;
	rat_factor_t rat_factors;
	knapsack_data_t knapdata;

	stage1_stat_init(&stats);
	stage1_bounds_init(&bounds, data);
	curr_poly_init(&currpoly, &bounds, data);
	rat_factor_init(&rat_factors, currpoly.gmp_N, 
			currpoly.gmp_a5, data->npr_in_p);
	knapsack_data_init(&knapdata, data->npr_in_p);

	search_a5(&currpoly, &rat_factors, &knapdata,
			&bounds, &stats);

	if (stats.n_survivors)
		success = 1;

	stage1_bounds_free(&bounds);
	curr_poly_free(&currpoly);
	rat_factor_free(&rat_factors);
	stage1_stat_free(&stats);
	knapsack_data_free(&knapdata);
	return success;
}
