#ifndef _STAGE1_H_
#define _STAGE1_H_

#include <poly_skew.h>

#define P0_MAX             46300	/* <=2^15.5 */
#define DEG 5
#define MULTIPLIER            60	/* 2*2*3*5 */
#define NPR5 50
#define HASHSHIFT  52 /*dont change this (or disable/adapt the asm-functions) */
#define NHASH   (1<<(64-HASHSHIFT))

#ifdef __cplusplus
extern "C" {
#endif

/*-----------------------------------------------------------------------*/

/* profiling and statistics */

typedef struct {
	uint64 n_a5;
	uint64 n_pr;
	uint64 n_p;
	uint64 n_p_p0;
	uint64 n_raw;
	uint64 n_polexpand;
	uint64 n_poltranslate;
	uint64 n_survivors;
} stage1_stat_t;

void stage1_stat_init(stage1_stat_t *stats);
void stage1_stat_free(stage1_stat_t *stats);

/*-----------------------------------------------------------------------*/
/* search bounds */

typedef struct {
	mpz_t gmp_a5_begin;
	mpz_t gmp_a5_end;
	double norm_max; 
	double skewness_min; 
	double skewness_max;
	double a3_max;
	double p_size_max;
	unsigned int p0_limit;
} bounds_t;

void stage1_bounds_init(bounds_t *bounds, poly_stage1_t *data);
void stage1_bounds_free(bounds_t *bounds);
void stage1_bounds_update(bounds_t *bounds, double N, double a5);

/*-----------------------------------------------------------------------*/

/* data for the current polynomial */

typedef struct {
	mpz_t gmp_N; 
	mpz_t gmp_prod; 
	mpz_t gmp_prod_base; 
	mpz_t gmp_root;
	mpz_t gmp_help1;
	mpz_t gmp_help2;
	mpz_t gmp_help3;
	mpz_t gmp_help4;
	mpz_t gmp_a5;
	mpz_t gmp_d;
	mpz_t gmp_m0;
	stage1_callback_t callback;
	void *callback_data;
} curr_poly_t;

void curr_poly_init(curr_poly_t *poly, bounds_t *bounds, 
			poly_stage1_t *data);
void curr_poly_free(curr_poly_t *poly);

/*-----------------------------------------------------------------------*/

/* data for the factors of the leading
   rational polynomial coefficient */

typedef struct {
	uint32 prime;
	uint32 sieve_start;
	uint32 sieve_step;
	double logprime;
	uint32 N_inv_mod_p;
	uint32 N_mod_p2;

	uint32 roots[DEG];
	uint32 my_idx;
	uint32 a5_mod_p2;
	uint32 minus5a5_mod_p;
} rat_prime_t;

typedef struct {
	uint32 p;
	uint32 r;
} aux_factor_t;

typedef struct {
	rat_prime_t primes[NPR5];
	unsigned char *bitarrays[NPR5];
	uint32 inv_table[NPR5][NPR5];

	uint32 npr_in_p;
	uint32 npr_total;
	uint32 a5_p_idx[NPR5];
	uint32 curr_p_idx[NPR5];

	uint32 p0_list_len;
	uint32 p0_list_len_max;
	uint32 last_p0;
	aux_factor_t *p0_list;
} rat_factor_t;

void rat_factor_init(rat_factor_t *r, mpz_t N, 
			mpz_t a5_start, uint32 npr_in_p);
void rat_factor_free(rat_factor_t *r);
uint32 rat_factor_find(rat_factor_t *rat_factors,
			mpz_t a5, double p_size_max);
void rat_factor_fill_aux(rat_factor_t *rat_factors, mpz_t N, mpz_t a5,
				double p0_max, uint32 p0_limit);

/*-----------------------------------------------------------------------*/
/* data for the knapsack code */

typedef struct {
	mpz_t gmp_disp;
	mpz_t gmp_D0;
	mpz_t gmp_D[NPR5][5];
	mpz_t gmp_kappa0;
	mpz_t gmp_kappa[NPR5][5];
	mpz_t gmp_kappa_help[NPR5 + 1];
	double dbl_kappa_help[NPR5 + 1];
	uint64_t ull_kappa0;
	uint64_t ull_kappa[NPR5][5];
	unsigned int ul_d_help[NPR5][5];

	uint64_t *s1;
	uint64_t *s2;
	uint64_t *s1sort;
	uint64_t *s2sort;
	int s1len;
	int s2len;
	uint64_t ull_bound;
	uint64_t lambda0;
	uint64_t *stored_pairs;
	int store_len;
	int nstore;
	unsigned int p_mod_p2[NPR5];
	
	int len1;
	int len2;
	int *sort;
	unsigned int *s1sortl, *s2sortl;
	uint64_t **hashptr;
} knapsack_data_t;

void knapsack_data_init(knapsack_data_t *knap, uint32 npr_in_p);
void knapsack_data_free(knapsack_data_t *knap);

/*-----------------------------------------------------------------------*/

void init_knapsack_exact(curr_poly_t *curr, knapsack_data_t *knap,
			rat_factor_t *rat_factors,
			double a3_max);
void init_knapsack_exact_p0(curr_poly_t *curr, knapsack_data_t *knap,
			rat_factor_t *rat_factors,
			unsigned int p0, unsigned int r0, double a3_max);
void knapsack_exact(curr_poly_t *curr, knapsack_data_t *knap,  
			bounds_t *bounds, stage1_stat_t *stats);

/*-----------------------------------------------------------------------*/
static INLINE void
mpz_get_ull_64(uint64_t * targptr, mpz_t src)
{
	if (src->_mp_size == 0)
		*targptr = 0;
	else if (src->_mp_size == 1)
		*targptr = (uint64_t) (src->_mp_d[0]);
	else {
		*targptr = (uint64_t) (src->_mp_d[1]);
		*targptr <<= 32;
		*targptr += (uint64_t) (src->_mp_d[0]);
	}
}

#ifdef __cplusplus
}
#endif

#endif /* !_STAGE1_H_ */
