#include "stage2.h"

#define LOCAL_BLOCK 40

typedef double (*objective_func)(double *v, void *extra);

/*----------------------------------------------------------------------*/
static uint32
stage2_root_score(int deg1, mpz_t *coeff1, 
		 uint32 prime_bound, double *score,
		 uint32 projective_only)
{
	uint32 i;
	mp_poly_t apoly;

	memset(&apoly, 0, sizeof(mp_poly_t));

	apoly.degree = deg1;
	for (i = 0; i <= deg1; i++) {
		gmp2mp(coeff1[i], &apoly.coeff[i].num);
		apoly.coeff[i].sign = POSITIVE;
		if (mpz_sgn(coeff1[i]) < 0)
			apoly.coeff[i].sign = NEGATIVE;
	}

	if (projective_only)
		return analyze_poly_roots_projective(&apoly, 
						prime_bound, score);
	else
		return analyze_poly_roots(&apoly, prime_bound, score);
}

/*-------------------------------------------------------------------------*/
static void
translate_dbl(double *c, double *d, double x)
{
	double d0 = d[0];
	double d1 = d[1];
	double d2 = d[2];
	double d3 = d[3];
	double d4 = d[4];
	double d5 = d[5];

	c[5] = d5;
	c[4] = d4 + x * (-5 * d5);
	c[3] = d3 + x * (-4 * d4 + x * 10 * d5);
	c[2] = d2 + x * (-3 * d3 + x * (6 * d4 - x * 10 * d5));
	c[1] = d1 + x * (-2 * d2 + x * (3 * d3 + x * (-4 * d4 + x * 5 * d5)));
	c[0] = d0 + x * (-d1 + x * (d2 + x * (-d3 + x * (d4 - x * d5))));
}

/*-------------------------------------------------------------------------*/
static void
translate_gmp(curr_poly_t *c, mpz_t * gmp_z, 
		mpz_t * lin, int k)
{
	mpz_set_si(c->gmp_help1, k);
	mpz_neg(c->gmp_help1, c->gmp_help1);	/* help1=-k */

	mpz_set(c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[1], c->gmp_help3);
	mpz_add(gmp_z[0], gmp_z[0], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[2], c->gmp_help3);
	mpz_add(gmp_z[0], gmp_z[0], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[3], c->gmp_help3);
	mpz_add(gmp_z[0], gmp_z[0], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[4], c->gmp_help3);
	mpz_add(gmp_z[0], gmp_z[0], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[5], c->gmp_help3);
	mpz_add(gmp_z[0], gmp_z[0], c->gmp_help2);
/* a0<-a0-a1*k+a2*k^2-a3*k^3+a4*k^4-a5*k^5 */

	mpz_set(c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[2], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 2);
	mpz_add(gmp_z[1], gmp_z[1], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[3], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 3);
	mpz_add(gmp_z[1], gmp_z[1], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[4], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 4);
	mpz_add(gmp_z[1], gmp_z[1], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[5], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 5);
	mpz_add(gmp_z[1], gmp_z[1], c->gmp_help2);
/* a1<-a1-2*a2*k+3*a3*k^2-4*a4*k^3+5*a5*k^4 */

	mpz_set(c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[3], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 3);
	mpz_add(gmp_z[2], gmp_z[2], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[4], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 6);
	mpz_add(gmp_z[2], gmp_z[2], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[5], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 10);
	mpz_add(gmp_z[2], gmp_z[2], c->gmp_help2);
/* a2<-a2-3*a3*k+6*a4*k^2-10*a5*k^3 */

	mpz_set(c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[4], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 4);
	mpz_add(gmp_z[3], gmp_z[3], c->gmp_help2);
	mpz_mul(c->gmp_help3, c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[5], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 10);
	mpz_add(gmp_z[3], gmp_z[3], c->gmp_help2);
/* a3<-a3-4*a4*k+10*a5*k^2 */

	mpz_set(c->gmp_help3, c->gmp_help1);
	mpz_mul(c->gmp_help2, gmp_z[5], c->gmp_help3);
	mpz_mul_ui(c->gmp_help2, c->gmp_help2, 5);
	mpz_add(gmp_z[4], gmp_z[4], c->gmp_help2);
/* a4<-a4-5*a5*k */

	mpz_mul(c->gmp_help2, lin[1], c->gmp_help1);
	mpz_add(lin[0], lin[0], c->gmp_help2);
/* lin0<-lin0-lin1*k */
}

/*-------------------------------------------------------------------------*/
enum min_varname {
	TRANSLATE_SIZE = 0,
	SKEWNESS,
	ROTATE0,
	ROTATE1,
	ROTATE2,
	MAX_VARS   /* must be last */
};

typedef struct {
	int num_vars;
	double initial_skew;
	dpoly_t *rpoly;
	dpoly_t *apoly;
} opt_data_t;

static double poly_callback(double *v, void *extra)
{
	opt_data_t *opt = (opt_data_t *)extra;
	double translated[MAX_POLY_DEGREE + 1];
	double t = floor(v[TRANSLATE_SIZE] + 0.5);
	double s = v[SKEWNESS];
	dpoly_t *apoly = opt->apoly;

	translate_dbl(translated, apoly->coeff, t);

	if (opt->num_vars > 2) {
		dpoly_t *rpoly = opt->rpoly;
		double r1 = rpoly->coeff[1];
		double r0 = rpoly->coeff[0] - t * r1;
		double c1 = floor(v[ROTATE1] + 0.5);
		double c0 = floor(v[ROTATE0] + 0.5);

		translated[0] +=           c0 * r0;
		translated[1] += c1 * r0 + c0 * r1;
		translated[2] += c1 * r1;
	}

	return ifs(translated, s);
}

/*-------------------------------------------------------------------------*/
#define PHI 1.618033988749
#define COMP_PHI (2.0 - PHI)

static double evaluate(double *base, double dist, 
			double *search_dir, uint32 ndim,
			objective_func callback, void *extra)
{
	uint32 i;
	double curr_pt[MAX_VARS];

	for (i = 0; i < ndim; i++)
		curr_pt[i] = base[i] + dist * search_dir[i];

	return callback(curr_pt, extra);
}

static void bracket_min(double *base, double *search_dir,
			double *a_out, double *b_out, double *c_out,
			double *fb_out, uint32 ndim, 
			objective_func callback, void *extra)
{
	double fa, fb, fc, t;
	double a = *a_out;
	double b = *b_out;
	double c;

	fa = evaluate(base, a, search_dir, ndim, callback, extra);
	fb = evaluate(base, b, search_dir, ndim, callback, extra);
	if (fb > fa) {
		t = a; a = b; b = t;
		t = fa; fa = fb; fb = t;
	}
	c = b + PHI * (b - a);
	fc = evaluate(base, c, search_dir, ndim, callback, extra);

	while (fb > fc) {
		double r = (b - a) * (fb - fc);
		double q = (b - c) * (fb - fa);
		double u, max_u, fu;

		t = fabs(q - r);
		t = MAX(t, 1e-20);
		if (q - r < 0.0)
			t = -t;

		u = b - ((b - c) * q - (b - a) * r) / (2.0 * t);
		max_u = b + 100.0 * (c - b);

		if ((b - u) * (u - c) > 0.0) {
			fu = evaluate(base, u, search_dir, ndim, 
					callback, extra);

			if (fu < fc) {
				*a_out = b; *b_out = u; *c_out = c;
				*fb_out = fu;
				return;
			}
			else if (fu > fb) {
				*a_out = a; *b_out = b; *c_out = u;
				*fb_out = fb;
				return;
			}

			u = c + PHI * (c - b);
			fu = evaluate(base, u, search_dir, ndim,
					callback, extra);
		}
		else if ((c - u) * (u - max_u) > 0.0) {
			fu = evaluate(base, u, search_dir, ndim,
					callback, extra);

			if (fu < fc) {
				b = c; c = u; u = c + PHI * (c - b);
				fb = fc; 
				fc = fu; 
				fu = evaluate(base, u, search_dir, ndim,
						callback, extra);
			}
		}
		else if ((u - max_u) * (max_u - c) >= 0.0) {
			u = max_u;
			fu = evaluate(base, u, search_dir, ndim,
					callback, extra);
		}
		else {
			u = c + PHI * (c - b);
			fu = evaluate(base, u, search_dir, ndim,
					callback, extra);
		}

		a = b; b = c; c = u;
		fa = fb; fb = fc; fc = fu;
	}

	*a_out = a; *b_out = b; *c_out = c;
	*fb_out = fb;
}

static double minimize_line_core(double *base, double *search_dir,
			double a_in, double b_in, double c_in, 
			double fb_in, double tol, double *min_out, 
			int32 ndim, objective_func callback, void *extra)
{
	int32 i;
	double a, b, x, w, v, u;
	double fx, fw, fv, fu;
	double e = 0.0;

	a = MIN(a_in, c_in);
	b = MAX(a_in, c_in);
	x = w = v = b_in;
	fx = fw = fv = fb_in;

	for (i = 0; i < 100; i++) {
		double xm = 0.5 * (a + b);
		double tol1 = tol * fabs(x) + 1e-10;
		double tol2 = 2.0 * tol1;
		double d = 0, etemp;

		if (fabs(x - xm) <= tol2 - 0.5 * (b - a)) {
			*min_out = x;
			return fx;
		}

		if (fabs(e) > tol1) {
			double r = (x - w) * (fx - fv);
			double q = (x - v) * (fx - fw);
			double p = (x - v) * q - (x - w) * r;
			
			q = 2.0 * (q - r);
			if (q > 0.0)
				p = -p;
			q = fabs(q);
			etemp = e;
			e = d;
			if (fabs(p) >= fabs(0.5 * q * etemp) ||
			    p <= q * (a - x) || p >= q * (b - x)) {
	
				e = b - x;
				if (x >= xm)
					e = a - x;				
				d = COMP_PHI * e;
			}
			else {
				d = p / q;
				u = x + d;
				if (u - a < tol2 || b - u < tol2)
					d = (xm >= x) ? tol1 : -tol1;
			}
		}
		else {
			e = (x >= xm) ? a - x : b - x;
			d = COMP_PHI * e;
		}

		if (fabs(d) < tol1)
			u = x + ((d >= 0) ? tol1 : -tol1);
		else
			u = x + d;
		fu = evaluate(base, u, search_dir, ndim, callback, extra);

		if (fu <= fx) {
			if (u >= x)
				a = x;
			else
				b = x;

			v = w; w = x; x = u;
			fv = fw; fw = fx; fx = fu;
		}
		else {
			if (u < x)
				a = u;
			else
				b = u;
			if (fu <= fw || w == x) {
				v = w; w = u;
				fv = fw; fw = fu;
			}
			else if (fu <= fv || v == x || v == w) {
				v = u;
				fv = fu;
			}
		}
	}

	printf("too many line iterations\n");
	*min_out = x;
	return fx;
}

static double minimize_line(double *base, double *search_dir, 
				uint32 ndim, objective_func callback,
				void *extra)
{
	uint32 i;
	double a = 0.0;
	double b = 1.0;
	double c, fb, min_dist, fmin_dist;

	bracket_min(base, search_dir, &a, &b, &c, 
			&fb, ndim, callback, extra);
	fmin_dist = minimize_line_core(base, search_dir, a, b, c, 
					fb, 1e-5, &min_dist, ndim,
					callback, extra);
	for (i = 0; i < ndim; i++) {
		search_dir[i] *= min_dist;
		base[i] += search_dir[i];
	}
	return fmin_dist;
}

static int32 optimize_local(double p[MAX_VARS], uint32 ndim, 
				double ftol, uint32 max_iter,
				objective_func callback, void *extra)
{
	uint32 i, j;
	double best_f, curr_f;
	double p_old[MAX_VARS], p_extrap[MAX_VARS], dir_extrap[MAX_VARS];
	double directions[MAX_VARS][MAX_VARS];

	memset(directions, 0, sizeof(directions));
	for (i = 0; i < ndim; i++)
		directions[i][i] = 1.0;
	best_f = callback(p, extra);

	for (i = 0; i < max_iter; i++) {

		int32 ibig = 0;
		double del = 0.0;
		double start_f = best_f;

		for (j = 0; j < ndim; j++) {
			curr_f = minimize_line(p, directions[j], ndim,
						callback, extra);
			if (best_f - curr_f > del) {
				ibig = j;
				del = best_f - curr_f;
			}
			best_f = curr_f;
		}

		if (2.0 * fabs(start_f - best_f) <= 
				ftol * (fabs(start_f) + fabs(best_f)) + 1e-25)
			return i + 1;

		for (j = 0; j < ndim; j++) {
			p_extrap[j] = 2.0 * p[j] - p_old[j];
			dir_extrap[j] = p[j] - p_old[j];
			p_old[j] = p[j];
		}

		curr_f = callback(p_extrap, extra);
		if (curr_f < start_f) {

			double t1 = start_f - best_f - del;
			double t2 = start_f - curr_f;
			double t = 2.0 * (start_f - 2.0 * best_f + curr_f) *
					t1 * t1 - del * t2 * t2;

			if (t < 0.0) {
				best_f = minimize_line(p, dir_extrap, ndim,
							callback, extra);
				for (j = 0; j < ndim; j++) {
					directions[ibig][j] = 
						directions[ndim-1][j];
					directions[ndim-1][j] = 
						dir_extrap[j];
				}
			}
		}
	}

	return i;
}

/*-------------------------------------------------------------------------*/
void
optimize_initial(poly_stage2_t *data,
		double *skewness, double *pol_norm,
		double *alpha_proj)
{
	stage2_curr_data_t *curr_data = 
			(stage2_curr_data_t *)(data->internal);
	curr_poly_t *c = &curr_data->curr_poly;
	opt_data_t opt_data;
	int i, j;
	double s, s0;
	double best[MAX_VARS];
	dpoly_t rpoly, apoly;

	opt_data.num_vars = 4;
	opt_data.rpoly = &rpoly;
	opt_data.apoly = &apoly;

	rpoly.degree = 1;
	rpoly.coeff[0] = mpz_get_d(c->gmp_lina[0]);
	rpoly.coeff[1] = mpz_get_d(c->gmp_lina[1]);
	apoly.degree = 5;
	for (i = 0; i < 6; i++)
		apoly.coeff[i] = mpz_get_d(c->gmp_a[i]);

	s = sqrt(fabs(apoly.coeff[2] / apoly.coeff[4]));
	s0 = fabs(apoly.coeff[2] / apoly.coeff[3]);
	if (s0 > s)
		s = s0;

	best[TRANSLATE_SIZE] = 0;
	best[SKEWNESS] = s;
	best[ROTATE0] = 0;
	best[ROTATE1] = 0;

	for (i = 0; i < 20; i++) {
		int32 iter = optimize_local(best, 4, 1e-5, LOCAL_BLOCK,
					poly_callback, &opt_data);

		translate_gmp(c, c->gmp_a, c->gmp_lina, 
				(int)floor(best[TRANSLATE_SIZE] + 0.5));

		mpz_set(c->gmp_p, c->gmp_lina[1]);
		mpz_neg(c->gmp_d, c->gmp_lina[0]);

		mpz_set_d(c->gmp_help1, floor(best[ROTATE1] + 0.5));
		mpz_mul(c->gmp_help2, c->gmp_help1, c->gmp_p);
		mpz_add(c->gmp_a[2], c->gmp_a[2], c->gmp_help2);
		mpz_mul(c->gmp_help1, c->gmp_help1, c->gmp_d);
		mpz_sub(c->gmp_a[1], c->gmp_a[1], c->gmp_help1);

		mpz_set_d(c->gmp_help1, floor(best[ROTATE0] + 0.5));
		mpz_mul(c->gmp_help2, c->gmp_help1, c->gmp_p);
		mpz_add(c->gmp_a[1], c->gmp_a[1], c->gmp_help2);
		mpz_mul(c->gmp_help1, c->gmp_help1, c->gmp_d);
		mpz_sub(c->gmp_a[0], c->gmp_a[0], c->gmp_help1);

		s = best[SKEWNESS];
		rpoly.coeff[1] = mpz_get_d(c->gmp_lina[1]);
		rpoly.coeff[0] = mpz_get_d(c->gmp_lina[0]);
		for (j = 0; j < 6; j++)
			apoly.coeff[j] = mpz_get_d(c->gmp_a[j]);

		if (iter < 2)
			break;

		best[TRANSLATE_SIZE] = 0;
		best[ROTATE0] = 0;
		best[ROTATE1] = 0;
	}

	*skewness = s;
	*pol_norm = sqrt(ifs(apoly.coeff, s));
	stage2_root_score(5, c->gmp_a, 100, alpha_proj, 1);

#if 0
	printf("norm %.7e\n", (*pol_norm) * (*pol_norm));
	for (i = 0; i < 2; i++)
		printf("%+.0lf\n", mpz_get_d(c->gmp_lina[i]));
	for (i = 0; i < 6; i++)
		printf("%+.0lf\n", mpz_get_d(c->gmp_a[i]));
#endif
}

/*-------------------------------------------------------------------------*/
static void
optimize_middle(curr_poly_t *c, double skewness, 
		double *new_skewness, double *norm_ptr)
{
	int i;
	opt_data_t opt_data;
	double best[MAX_VARS];
	dpoly_t rpoly, apoly;

	opt_data.num_vars = 2;
	opt_data.rpoly = &rpoly;
	opt_data.apoly = &apoly;

	rpoly.degree = 1;
	rpoly.coeff[0] = mpz_get_d(c->gmp_linb[0]);
	rpoly.coeff[1] = mpz_get_d(c->gmp_linb[1]);
	apoly.degree = 5;
	for (i = 0; i < 6; i++)
		apoly.coeff[i] = mpz_get_d(c->gmp_b[i]);

	best[TRANSLATE_SIZE] = 0;
	best[SKEWNESS] = skewness;

	optimize_local(best, 2, 1e-7, LOCAL_BLOCK,
			poly_callback, &opt_data);

	translate_gmp(c, c->gmp_b, c->gmp_linb, 
			(int)(best[TRANSLATE_SIZE] + 0.5));

	for (i = 0; i < 6; i++)
		apoly.coeff[i] = mpz_get_d(c->gmp_b[i]);

	*new_skewness = best[SKEWNESS];
	*norm_ptr = sqrt(ifs(apoly.coeff, *new_skewness));
}

/*-------------------------------------------------------------------------*/
static void
optimize_final_core(curr_poly_t *c, assess_t *assess,
			double *eptr, double root_score)
{
	int i;
	double size_score;
	dpoly_t rpoly, apoly;

	root_score = pow(exp(root_score), -2./6);

	rpoly.degree = 1;
	for (i = 0; i < 2; i++)
		rpoly.coeff[i] = mpz_get_d(c->gmp_linb[i]);

	apoly.degree = 5;
	for (i = 0; i < 6; i++)
		apoly.coeff[i] = mpz_get_d(c->gmp_b[i]);

	analyze_poly_size(&assess->integ_aux, 
			&rpoly, &apoly, &size_score);
	if (size_score == 0.0)
		printf("error: size score computation failed\n");

	*eptr = size_score * root_score;
}

/*-------------------------------------------------------------------------*/
void
optimize_final(int x, int y, poly_stage2_t *data, double skewness)
{
	int i;
	double alpha, norm, combined_score;
	stage2_curr_data_t *s = (stage2_curr_data_t *)data->internal;
	curr_poly_t *c = &s->curr_poly;
	assess_t *assess = &s->assess;

	for (i = 0; i < 6; i++)
		mpz_set(c->gmp_b[i], c->gmp_a[i]);
	mpz_set_si(c->gmp_help1, y);
	mpz_mul(c->gmp_help2, c->gmp_help1, c->gmp_p);
	mpz_add(c->gmp_b[2], c->gmp_b[2], c->gmp_help2);
	mpz_mul(c->gmp_help1, c->gmp_help1, c->gmp_d);
	mpz_sub(c->gmp_b[1], c->gmp_b[1], c->gmp_help1);
	mpz_set_si(c->gmp_help1, x);
	mpz_mul(c->gmp_help2, c->gmp_help1, c->gmp_p);
	mpz_add(c->gmp_b[1], c->gmp_b[1], c->gmp_help2);
	mpz_mul(c->gmp_help1, c->gmp_help1, c->gmp_d);
	mpz_sub(c->gmp_b[0], c->gmp_b[0], c->gmp_help1);
	for (i = 0; i < 2; i++)
		mpz_set(c->gmp_linb[i], c->gmp_lina[i]);

	if (stage2_root_score(5, c->gmp_b, data->p_bound, &alpha, 0))
		return;

	if (alpha > -4.5)
		return;

	optimize_final_core(c, assess, &combined_score, alpha);

	if (combined_score >= data->min_e) {
		optimize_middle(c, skewness, &skewness, &norm);
		data->callback(data->callback_data, 5, c->gmp_b, 
				c->gmp_linb, skewness, norm, alpha, 
				combined_score);
	}
}
