#include "stage2.h"

#define NUM_ROOT_HEAP_ENTRIES 400
#define MAX_LIST_LEN 10000
#define MAX_SIEVE_PRIME      100
#define MAX_SIEVE_PRIME_POWER 1000
#define LOG_SCALE_FACTOR 1000
#define UNROLL 4
#define MAX_X       1000000	/* !!! */
#define DEFAULT_BLOCK_SIZE  8192

/*-------------------------------------------------------------------------*/
#define COMPUTE_IFS   s2=sc; res=0.; \
  res+=s2*sq[3]/35.; res+=sq[2]/s2/35.;   \
  s2*=(sc*sc);   \
  res+=s2*sq[4]/27.; res+=sq[1]/s2/27.;   \
  s2*=(sc*sc);   \
  res+=s2*sq[5]/11.; res+=sq[0]/s2/11.


static double
find_best_skewness(double *coeff, double s0)
{				/* degree=5 */
	double sq[6], d, s, sc, s2, res, ds, v;
	int i, j, k;

	for (i = 0; i < 6; i++)
		sq[i] = coeff[i] * coeff[i];
	for (i = 0; i < 4; i++) {
		d = 2 * coeff[i];
		k = i + 1;
		for (j = i + 2; j < 6; j += 2, k++)
			sq[k] += d * coeff[j];
	}
	ds = 10.;
	s = s0;
	sc = s;
	COMPUTE_IFS;
	v = res;
	while (ds > 2.) {
		sc = s + ds;
		COMPUTE_IFS;
		if (res < v) {
			s = sc;
			v = res;
			ds *= 1.1;
			continue;
		}
		if (ds < s) {
			sc = s - ds;
			COMPUTE_IFS;
			if (res < v) {
				s = sc;
				v = res;
				ds *= 1.1;
				continue;
			}
		}
		ds *= 0.9;
	}
	return s;
}

/*-------------------------------------------------------------------------*/
static void
initsieve(curr_poly_t *c, root_sieve_t *rs, double sieve_bias)
{
	int i, j, k, len;
	uint32 p, pk, fi, mi, md, mp, step;
	uint32 co[6], difftable[6], fis[6];
	double dp = 0, dpk, dlp = 0, dlog;
	primelist_t *pl = rs->pl;

	k = p = 0;
	pk = MAX_SIEVE_PRIME_POWER;

	for (len = 0; len < MAX_LIST_LEN; pk *= p) {

		if (pk >= MAX_SIEVE_PRIME_POWER) {
			p += prime_delta[k++];
			if (p >= MAX_SIEVE_PRIME)
				break;
			dp = (double) p;
			dlp = log(dp);
			pk = p;
		}

		dpk = (double) pk;
		dlog = dlp / dpk * dp / (dp + 1.);
		for (i = 0; i <= 5; i++)
			co[i] = mpz_fdiv_ui(c->gmp_a[i], pk);
		md = mpz_fdiv_ui(c->gmp_d, pk);
		mp = mpz_fdiv_ui(c->gmp_p, pk);

		for (i = 0; i < 6; i++) {
			fi = co[5];
			for (j = 1; j <= 5; j++)
				fi = (fi * i + co[5 - j]) % pk;
			fis[i] = fi;
		}
		if (pk >= 6) {
			for (i = 0; i < 6; i++) {
				difftable[i] = fis[0];
				for (j = 0; j < 5 - i; j++) {
					fis[j] = mp_modsub_1(fis[j + 1], 
								fis[j], pk);
				}
			}
		}

		for (i = 0; i < pk; i++) {
			mi = mp_modsub_1(md, mp_modmul_1(i, mp, pk), pk);

			if (pk < 6) {
				fi = fis[i];
			}
			else {
				fi = difftable[0];
				for (j = 0; j < 5; j++) {
					difftable[j] = mp_modadd_1(difftable[j],
								difftable[j+1],
								pk);
				}
			}

			if (mi != 0) {
				step = pk;
				while (mi % p == 0 && fi % p == 0) {
					mi /= p;
					fi /= p;
					step /= p;
				}
				if (mi % p) {
					pl[len].pk = pk;
					pl[len].base_p = p;
					pl[len].resclass = i;
					pl[len].step = step;
					pl[len].logpk = (uint16)(0.5 + dlog *
							LOG_SCALE_FACTOR);
					if (pl[len].logpk == 0)
						pl[len].logpk = 1;

					pl[len].start = mp_modmul_1(fi, 
							mp_modinv_1(mi, step), 
							step);
					len++;
				}	/* otherwise no solutions */
			}
			else if (fi == 0) {
				/* p^k | fi+j*(i-m) for all j */
				sieve_bias += dlog;
			}
		}
	}

	rs->primelistlen = len;
	rs->sieve_bias = sieve_bias;
}

/*-------------------------------------------------------------------------*/
static void
compute_line_size(double max_norm, double *dbl_a, 
		  double dbl_p, double dbl_d, 
		double skewness, int *xmin, int *xmax)
{
	int x0, x1, xx;
	double da0, da1;
	double v0;

	da0 = dbl_a[0];
	da1 = dbl_a[1];
	max_norm = max_norm * max_norm;

	x0 = 0;
	x1 = MAX_X;
	dbl_a[1] = da1 - ((double) x1) * dbl_p;
	dbl_a[0] = da0 + ((double) x1) * dbl_d;
	v0 = ifs(dbl_a, find_best_skewness(dbl_a, skewness));
	if (v0 > max_norm) {
		while (x1 - x0 > 1) {
			xx = (x0 + x1) / 2;
			dbl_a[1] = da1 - ((double) xx) * dbl_p;
			dbl_a[0] = da0 + ((double) xx) * dbl_d;
			v0 = ifs(dbl_a, find_best_skewness(dbl_a, skewness));
			if (v0 > max_norm)
				x1 = xx;
			else
				x0 = xx;
		}
	}
	*xmin = -x1;

	x0 = 0;
	x1 = MAX_X;
	dbl_a[1] = da1 + ((double) x1) * dbl_p;
	dbl_a[0] = da0 - ((double) x1) * dbl_d;
	v0 = ifs(dbl_a, find_best_skewness(dbl_a, skewness));
	if (v0 > max_norm) {
		while (x1 - x0 > 1) {
			xx = (x0 + x1) / 2;
			dbl_a[1] = da1 + ((double) xx) * dbl_p;
			dbl_a[0] = da0 - ((double) xx) * dbl_d;
			v0 = ifs(dbl_a, find_best_skewness(dbl_a, skewness));
			if (v0 > max_norm)
				x1 = xx;
			else
				x0 = xx;
		}
	}
	*xmax = x1;

	dbl_a[0] = da0;
	dbl_a[1] = da1;
}

/*-------------------------------------------------------------------------*/
static void
compute_sieve_params(int *xmin, int *xmax,
			int *block_size_out,
			int *num_blocks_out)
{
	int extra;
	int sieve_block_size;
	int num_sieve_blocks;
	int sieve_size, x_range, x_first;

	sieve_block_size = DEFAULT_BLOCK_SIZE;
	x_range = *xmax - *xmin;
	if (2 * x_range < sieve_block_size)
		sieve_block_size /= 2;
	num_sieve_blocks = (x_range + (sieve_block_size - 1)) / 
				sieve_block_size;
	sieve_size = num_sieve_blocks * sieve_block_size;
	extra = (sieve_size - x_range) / 2;
	x_first = *xmin - extra;
	*xmin = x_first;
	*xmax = x_first + sieve_size - 1;
	*block_size_out = sieve_block_size;
	*num_blocks_out = num_sieve_blocks;
}

/*-------------------------------------------------------------------------*/
static void
prepare_sieve(root_sieve_t *rs, int xmin, int y)
{
	uint16 *array;
	primelist_t *pl = rs->pl;
	pk_array_t *pk_array = rs->pk_arrays;
	int primelistlen = rs->primelistlen;
	int num_pk_array = 0;
	int i, i0, j, pk;

	for (i0 = 0, i = 1; i <= primelistlen; i++) {

		if (i < primelistlen && pl[i].base_p == pl[i0].base_p)
			continue;

		pk = pl[i-1].pk;
		array = pk_array->array_ptr;
		memset(array, 0, pk * sizeof(uint16));

		for (j = i0; j < i; j++) {
			int32 curr = pl[j].start;
			int32 step = pl[j].step;
			uint16 logpk = pl[j].logpk;

			curr = (curr - y * pl[j].resclass - xmin) % step;
			if (curr < 0)
				curr += step;

			while (curr < pk) {
				array[curr] += logpk;
				curr += step;
			}
		}
		for (j = 1; j < UNROLL; j++)
			memcpy(array + j * pk, array, pk * sizeof(uint16));

		pk_array->curr_offset = 0;
		pk_array->pk_array_len = pk;
		pk_array++;
		num_pk_array++;
		i0 = i;
	}
	rs->curr_num_pk_array = num_pk_array;
}

/*-------------------------------------------------------------------------*/
static void
sieve_one_block(root_sieve_t *rs, int sieve_block_size)
{
	int i, j;
	uint64 *block_array = (uint64 *)rs->sieve_block;
	int block_words = sieve_block_size / UNROLL;

	memset(block_array, 0, block_words * sizeof(uint64));

	for (i = 0; i < rs->curr_num_pk_array; i++) {
		pk_array_t *pk = rs->pk_arrays + i;
		uint64 *pk_array = (uint64 *)pk->array_ptr;
		int pk_words = pk->pk_array_len;
		int pk_offset = pk->curr_offset;
		int block_offset = 0;

		while (block_offset < block_words) {
			int curr_words = MIN(pk_words - pk_offset,
					     block_words - block_offset);
			uint64 *b = block_array + block_offset;
			uint64 *p = pk_array + pk_offset;

#if (defined(__GNUC__) || defined(__ICL)) && \
	defined(__i386__) && defined(HAS_MMX)
			j = 0;
			asm volatile(
			    "cmpl $0, %3              \n\t"
			    "je 1f                   \n\t"
			    "0:                       \n\t"
			    "movq (%1,%0,8), %%mm0    \n\t"
			    "movq 8(%1,%0,8), %%mm1   \n\t"
			    "movq 16(%1,%0,8), %%mm2  \n\t"
			    "movq 24(%1,%0,8), %%mm3  \n\t"
			    "paddw (%2,%0,8), %%mm0   \n\t"
			    "paddw 8(%2,%0,8), %%mm1  \n\t"
			    "paddw 16(%2,%0,8), %%mm2 \n\t"
			    "paddw 24(%2,%0,8), %%mm3 \n\t"
			    "movq %%mm0, (%1,%0,8)    \n\t"
			    "movq %%mm1, 8(%1,%0,8)   \n\t"
			    "movq %%mm2, 16(%1,%0,8)  \n\t"
			    "movq %%mm3, 24(%1,%0,8)  \n\t"
			    "addl $4, %0              \n\t"
			    "cmpl %3, %0              \n\t"
			    "jne 0b                   \n\t"
			    "1:                       \n\t"
			    :"+r"(j)
			    :"r"(b), "r"(p), "g"(curr_words & ~3)
			    :"%mm0", "%mm1", "%mm2", "%mm3", 
			     "%xmm0", "%xmm1", "memory", "cc");
#else
			for (j = 0; j < (curr_words & ~3); j += 4) {
				b[j+0] += p[j+0];
				b[j+1] += p[j+1];
				b[j+2] += p[j+2];
				b[j+3] += p[j+3];
			}
#endif
			for (; j < curr_words; j++)
				b[j] += p[j];

			block_offset += curr_words;
			pk_offset += curr_words;
			if (pk_offset >= pk_words)
				pk_offset -= pk_words;
		}

		pk->curr_offset = pk_offset;
	}

#if (defined(__GNUC__) || defined(__ICL)) && \
	defined(__i386__) && defined(HAS_MMX)
	asm volatile("emms");
#endif
}

/*-------------------------------------------------------------------------*/
/* boilerplate code for managing a heap of polynomial rotations */

#define HEAP_SWAP(a,b) { tmp = a; a = b; b = tmp; }
#define HEAP_PARENT(i)  (((i)-1) >> 1)
#define HEAP_LEFT(i)    (2 * (i) + 1)
#define HEAP_RIGHT(i)   (2 * (i) + 2)

static void heapify(root_heap_t *h, uint32 index, uint32 size) {

	uint32 c;
	root_heap_t tmp;
	for (c = HEAP_LEFT(index); c < (size-1); 
			index = c, c = HEAP_LEFT(index)) {

		if( h[c].score > h[c+1].score )
			c++;

		if( h[index].score > h[c].score ) {
			HEAP_SWAP(h[index], h[c]);
		}
		else
			return;
	}
	if (c == (size-1) && h[index].score > h[c].score) {
		HEAP_SWAP(h[index], h[c]);
	}
}

static void make_heap(root_heap_t *h, uint32 size) {

	int32 i;
	for (i = HEAP_PARENT(size); i >= 0; i--)
		heapify(h, (uint32)i, size);
}

static void
save_rotation(root_sieve_t *rs, int x, int y, uint16 score)
{
	root_heap_t *h = rs->root_heap;
	if (rs->num_heap_entries <= NUM_ROOT_HEAP_ENTRIES - 1) {
		root_heap_t *rh = h + rs->num_heap_entries++;
		rh->x = x;
		rh->y = y;
		rh->score = score;
		if (rs->num_heap_entries == NUM_ROOT_HEAP_ENTRIES)
			make_heap(h, NUM_ROOT_HEAP_ENTRIES);
	}
	else if (h[0].score < score) {
		h[0].x = x;
		h[0].y = y;
		h[0].score = score;
		heapify(h, 0, NUM_ROOT_HEAP_ENTRIES);
	}
}

/*-------------------------------------------------------------------------*/
static void
root_sieve_xy(poly_stage2_t *data, double max_norm,
		double *a, double p, double d,
		double sk, int y0, int y_inc)
{
	int i, j;
	double a1 = a[1];
        double a2 = a[2];
	int xmin, xmax;
	int block_size, num_blocks;

	stage2_curr_data_t *s = (stage2_curr_data_t *)data->internal;
	root_sieve_t *rs = &s->root_sieve;

	while (1) {
		a[1] = a1 - y0 * d;
		a[2] = a2 + y0 * p;
		sk = find_best_skewness(a, sk);
		compute_line_size(max_norm, a, p, d, sk, &xmin, &xmax);
		if (xmax - xmin < 1000)
			break;

		compute_sieve_params(&xmin, &xmax, &block_size, &num_blocks);
		prepare_sieve(rs, xmin, y0);

		for (i = 0; i < num_blocks; i++) {
			sieve_one_block(rs, block_size);
			for (j = 0; j < block_size; j++) {
				save_rotation(rs, xmin + j, y0, 
						rs->sieve_block[j]);
			}
			xmin += block_size;
		}
		y0 += y_inc;
	}

	a[1] = a1;
	a[2] = a2;
}

/*-------------------------------------------------------------------------*/
void
root_sieve_run(poly_stage2_t *data, 
		double skewness, 
		double alpha_proj)
{
	int i;
	double dbl_p;
	double dbl_d;
	double dbl_sv[6];
	double max_norm;

	stage2_curr_data_t *s = (stage2_curr_data_t *)data->internal;
	curr_poly_t *c = &s->curr_poly;
	root_sieve_t *rs = &s->root_sieve;

	dbl_p = mpz_get_d(c->gmp_p);
	dbl_d = mpz_get_d(c->gmp_d);
	for (i = 0; i <= 5; i++)
		dbl_sv[i] = mpz_get_d(c->gmp_a[i]);

	max_norm = data->max_norm_1 * exp(-alpha_proj);
	initsieve(c, rs, -alpha_proj);

	rs->num_heap_entries = 0;

	root_sieve_xy(data, max_norm, 
			  dbl_sv, dbl_p, dbl_d, 
			  skewness, 0, 1);
	root_sieve_xy(data, max_norm, 
			  dbl_sv, dbl_p, dbl_d, 
			  skewness, -1, -1);

	for (i = 0; i < rs->num_heap_entries; i++) {
		if (rs->random_root_score - 
			(rs->sieve_bias + 
			 (double)rs->root_heap[i].score / 
			 LOG_SCALE_FACTOR) >= -4.0)
			continue;

		optimize_final(rs->root_heap[i].x, 
				rs->root_heap[i].y,
				data, skewness);
	}
}

/*-------------------------------------------------------------------------*/
void root_sieve_init(root_sieve_t *rs)
{
	int i, p;
	memset(rs, 0, sizeof(root_sieve_t));

	rs->pl = (primelist_t *)xmalloc(MAX_LIST_LEN * sizeof(primelist_t));
	rs->root_heap = (root_heap_t *)xmalloc(NUM_ROOT_HEAP_ENTRIES * 
						sizeof(root_heap_t));
	rs->sieve_block = (uint16 *)aligned_malloc(DEFAULT_BLOCK_SIZE *
						sizeof(uint16), 64);

	for (i = p = 0; i < PRECOMPUTED_NUM_PRIMES; i++) {
		p += prime_delta[i];
		if (p > MAX_SIEVE_PRIME)
			break;

		rs->random_root_score += log((double)p) / (p - 1);
	}
	rs->pk_arrays = (pk_array_t *)xmalloc(i * sizeof(pk_array_t));
	rs->num_pk_array = i;
	for (i = 0; i < rs->num_pk_array; i++) {
		rs->pk_arrays[i].array_ptr = (uint16 *)aligned_malloc(
						MAX_SIEVE_PRIME_POWER * 
						 UNROLL * sizeof(uint16), 64);
	}
}

/*-------------------------------------------------------------------------*/
void root_sieve_free(root_sieve_t *rs)
{
	int i;

	free(rs->pl);
	free(rs->root_heap);
	aligned_free(rs->sieve_block);
	for (i = 0; i < rs->num_pk_array; i++)
		aligned_free(rs->pk_arrays[i].array_ptr);
	free(rs->pk_arrays);
	memset(rs, 0, sizeof(root_sieve_t));
}
