#include "stage2.h"

/*----------------------------------------------------------------------*/
void
assess_init(assess_t *a, double b0, double b1, double area)
{
	a->bound0 = b0;
	a->bound1 = b1;
	a->area = area;
	integrate_init(&a->integ_aux, 1e-100, 0.01 * SIZE_EPS,
			double_exponential, 8);
}

/*----------------------------------------------------------------------*/
void
assess_free(assess_t *a)
{
	integrate_free(&a->integ_aux);
}

/*-------------------------------------------------------------------------*/
static int
pol_expand(curr_poly_t *c, mpz_t gmp_N)
{
	/* compute coefficients */
	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help4);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_a[5]);
	mpz_sub(c->gmp_help3, gmp_N, c->gmp_help4);
	mpz_fdiv_qr(c->gmp_help3, c->gmp_help1, c->gmp_help3, c->gmp_p);
	if (mpz_sgn(c->gmp_help1))
		return 0;

	if (mpz_cmp_ui(c->gmp_p, 1) == 0)
		mpz_set_ui(c->gmp_help2, 1);
	else {
		if (!mpz_invert(c->gmp_help2, c->gmp_d, c->gmp_p))
			return 0;
	}
	mpz_mul(c->gmp_help4, c->gmp_help2, c->gmp_help2);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help4);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help3);
	mpz_fdiv_r(c->gmp_a[4], c->gmp_help4, c->gmp_p);
	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help4);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_a[4]);
	mpz_sub(c->gmp_help3, c->gmp_help3, c->gmp_help4);
	mpz_fdiv_qr(c->gmp_help3, c->gmp_help1, c->gmp_help3, c->gmp_p);
	if (mpz_sgn(c->gmp_help1))
		return 0;

	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_d);
	mpz_fdiv_q(c->gmp_a[3], c->gmp_help3, c->gmp_help4);
	mpz_fdiv_q(c->gmp_a[3], c->gmp_a[3], c->gmp_p);
	mpz_mul(c->gmp_a[3], c->gmp_a[3], c->gmp_p);
	mpz_mul(c->gmp_help4, c->gmp_help2, c->gmp_help2);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help2);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help3);
	mpz_fdiv_r(c->gmp_help1, c->gmp_help4, c->gmp_p);
	mpz_add(c->gmp_a[3], c->gmp_a[3], c->gmp_help1);
	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_a[3]);
	mpz_sub(c->gmp_help3, c->gmp_help3, c->gmp_help4);
	mpz_fdiv_qr(c->gmp_help3, c->gmp_help1, c->gmp_help3, c->gmp_p);
	if (mpz_sgn(c->gmp_help1))
		return 0;

	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_d);
	mpz_fdiv_q(c->gmp_a[2], c->gmp_help3, c->gmp_help4);
	mpz_fdiv_q(c->gmp_a[2], c->gmp_a[2], c->gmp_p);
	mpz_mul(c->gmp_a[2], c->gmp_a[2], c->gmp_p);
	mpz_mul(c->gmp_help4, c->gmp_help2, c->gmp_help2);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_help3);
	mpz_fdiv_r(c->gmp_help1, c->gmp_help4, c->gmp_p);
	mpz_add(c->gmp_a[2], c->gmp_a[2], c->gmp_help1);
	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_d);
	mpz_mul(c->gmp_help4, c->gmp_help4, c->gmp_a[2]);
	mpz_sub(c->gmp_help3, c->gmp_help3, c->gmp_help4);
	mpz_fdiv_qr(c->gmp_help3, c->gmp_help1, c->gmp_help3, c->gmp_p);
	if (mpz_sgn(c->gmp_help1))
		return 0;

	mpz_fdiv_q(c->gmp_a[1], c->gmp_help3, c->gmp_d);
	mpz_fdiv_q(c->gmp_a[1], c->gmp_a[1], c->gmp_p);
	mpz_mul(c->gmp_a[1], c->gmp_a[1], c->gmp_p);
	mpz_mul(c->gmp_help4, c->gmp_help3, c->gmp_help2);
	mpz_fdiv_r(c->gmp_help1, c->gmp_help4, c->gmp_p);
	mpz_add(c->gmp_a[1], c->gmp_a[1], c->gmp_help1);
	mpz_mul(c->gmp_help4, c->gmp_d, c->gmp_a[1]);
	mpz_sub(c->gmp_help3, c->gmp_help3, c->gmp_help4);
	mpz_fdiv_qr(c->gmp_help3, c->gmp_help1, c->gmp_help3, c->gmp_p);
	if (mpz_sgn(c->gmp_help1))
		return 0;

	mpz_set(c->gmp_a[0], c->gmp_help3);

	mpz_fdiv_qr(c->gmp_help1, c->gmp_a[3], c->gmp_a[3], c->gmp_d);
	mpz_add(c->gmp_help2, c->gmp_a[3], c->gmp_a[3]);
	if (mpz_cmp(c->gmp_d, c->gmp_help2) < 0) {
		mpz_sub(c->gmp_a[3], c->gmp_a[3], c->gmp_d);
		mpz_add_ui(c->gmp_help1, c->gmp_help1, 1);
	}
	mpz_mul(c->gmp_help1, c->gmp_help1, c->gmp_p);
	mpz_add(c->gmp_a[4], c->gmp_a[4], c->gmp_help1);

	mpz_fdiv_qr(c->gmp_help1, c->gmp_a[2], c->gmp_a[2], c->gmp_d);
	mpz_add(c->gmp_help2, c->gmp_a[2], c->gmp_a[2]);
	if (mpz_cmp(c->gmp_d, c->gmp_help2) < 0) {
		mpz_sub(c->gmp_a[2], c->gmp_a[2], c->gmp_d);
		mpz_add_ui(c->gmp_help1, c->gmp_help1, 1);
	}
	mpz_mul(c->gmp_help1, c->gmp_help1, c->gmp_p);
	mpz_add(c->gmp_a[3], c->gmp_a[3], c->gmp_help1);

	mpz_set(c->gmp_lina[1], c->gmp_p);
	mpz_neg(c->gmp_lina[0], c->gmp_d);

	return 1;
}

/*-------------------------------------------------------------------------*/
static void
curr_poly_init(curr_poly_t *c)
{
	int i;

	mpz_init(c->gmp_p);
	mpz_init(c->gmp_d);
	for (i = 0; i < 2; i++)
		mpz_init(c->gmp_lina[i]);
	for (i = 0; i < 2; i++)
		mpz_init(c->gmp_linb[i]);
	for (i = 0; i < 6; i++)
		mpz_init(c->gmp_a[i]);
	for (i = 0; i < 6; i++)
		mpz_init(c->gmp_b[i]);
	mpz_init(c->gmp_help1);
	mpz_init(c->gmp_help2);
	mpz_init(c->gmp_help3);
	mpz_init(c->gmp_help4);
}

/*-------------------------------------------------------------------------*/
static void
curr_poly_free(curr_poly_t *c)
{
	int i;

	mpz_clear(c->gmp_p);
	mpz_clear(c->gmp_d);
	for (i = 0; i < 2; i++)
		mpz_clear(c->gmp_lina[i]);
	for (i = 0; i < 2; i++)
		mpz_clear(c->gmp_linb[i]);
	for (i = 0; i < 6; i++)
		mpz_clear(c->gmp_a[i]);
	for (i = 0; i < 6; i++)
		mpz_clear(c->gmp_b[i]);
	mpz_clear(c->gmp_help1);
	mpz_clear(c->gmp_help2);
	mpz_clear(c->gmp_help3);
	mpz_clear(c->gmp_help4);
}

/*-------------------------------------------------------------------------*/
void
poly_stage2_init(poly_stage2_t *data, 
		 stage2_callback_t callback,
		 void *callback_data)
{
	memset(data, 0, sizeof(poly_stage2_t));
	mpz_init(data->gmp_N);
	data->max_norm_1 = 1e20;
	data->max_norm_2 = 1e18;
	data->min_e = 0;
	data->p_bound = 2000;
	data->bound0 = 1e7;
	data->bound1 = 5e6;
	data->area = 1e16;

	data->callback = callback;
	data->callback_data = callback_data;
}

/*-------------------------------------------------------------------------*/
void
poly_stage2_free(poly_stage2_t *data)
{
	mpz_clear(data->gmp_N);

	if (data->internal) {
		stage2_curr_data_t *s = (stage2_curr_data_t *)(data->internal);
		curr_poly_free(&s->curr_poly);
		root_sieve_free(&s->root_sieve);
		assess_free(&s->assess);

		free(data->internal);
		data->internal = NULL;
	}
}

/*-------------------------------------------------------------------------*/
void
poly_stage2_run(poly_stage2_t *data, mpz_t a5, mpz_t p, mpz_t d)
{
	double skewness;
	double pol_norm;
	double alpha_proj;
	stage2_curr_data_t *s;

	s = (stage2_curr_data_t *)(data->internal);
	if (s == NULL) {
		s = (stage2_curr_data_t *)xmalloc(sizeof(stage2_curr_data_t));
		curr_poly_init(&s->curr_poly);
		root_sieve_init(&s->root_sieve);
		assess_init(&s->assess, data->bound0, 
				data->bound1, data->area);
		data->internal = (void *)s;
	}

	mpz_set(s->curr_poly.gmp_a[5], a5);
	mpz_set(s->curr_poly.gmp_p, p);
	mpz_set(s->curr_poly.gmp_d, d);
	if (!pol_expand(&s->curr_poly, data->gmp_N)) {
		fprintf(stderr, "expand failed\n");
		return;
	}

	optimize_initial(data, &skewness, &pol_norm, &alpha_proj);

	if (pol_norm * exp(alpha_proj) > data->max_norm_1)
		return;
	root_sieve_run(data, skewness, alpha_proj);
}
