#ifndef _STAGE2_H_
#define _STAGE2_H_

#include <poly_skew.h>

#ifdef __cplusplus
extern "C" {
#endif

/*-----------------------------------------------------------------------*/
/* data used in the current polynomial */

typedef struct {
	mpz_t gmp_a[6];
	mpz_t gmp_b[6];
	mpz_t gmp_lina[2];
	mpz_t gmp_linb[2];
	mpz_t gmp_help1;
	mpz_t gmp_help2;
	mpz_t gmp_help3;
	mpz_t gmp_help4;
	mpz_t gmp_p;
	mpz_t gmp_d;
} curr_poly_t;

/*-----------------------------------------------------------------------*/
/* data for rating polynomial yield */

typedef struct {
	double bound0;
	double bound1;
	double area;
	integrate_t integ_aux;
} assess_t;

void assess_init(assess_t *a, double b0, double b1, double area);

void assess_free(assess_t *a);

/*-----------------------------------------------------------------------*/
/* routines for optimizing polynomials */

void optimize_initial(poly_stage2_t *data, double *skewness, 
			double *pol_norm, double *alpha_proj);

void optimize_final(int x, int y, poly_stage2_t *data, double skewness);

/*-----------------------------------------------------------------------*/
/* data for the root sieve */

typedef struct {
	int pk;			/* p^k */
	int base_p;		/* p^1 */
	int resclass;		/* residue class mod p^k of this entry */
	int start;		/* sievearray[start] has zero =i mod p^k */
	int step;		/* stepwidth p^(k-l) */
	uint16 logpk;		/* scaled value */
} primelist_t;

typedef struct {
	uint16 *array_ptr;
	int curr_offset;
	int pk_array_len;
} pk_array_t;

typedef struct {
	int x;
	int y;
	uint16 score;
} root_heap_t;

typedef struct {
	primelist_t *pl;
	int primelistlen;
	double sieve_bias; 
	double random_root_score;

	uint16 *sieve_block;

	int num_pk_array; 
	int curr_num_pk_array;
	pk_array_t *pk_arrays;

	int num_heap_entries;
	root_heap_t *root_heap;
} root_sieve_t;

void root_sieve_init(root_sieve_t *rs);
void root_sieve_free(root_sieve_t *rs);
void root_sieve_run(poly_stage2_t *data, 
			double skewness, 
			double alpha_proj);

/*-------------------------------------------------------------------------*/

/* data for optimizing a single (a5, p, d) triplet */

typedef struct {
	curr_poly_t curr_poly;
	root_sieve_t root_sieve;
	assess_t assess;
} stage2_curr_data_t;

/*-------------------------------------------------------------------------*/
static INLINE double
ifs(double *coeff, double skewness)
{				/* degree=5 */
	double sq[6], d, s, res;
	int i, j, k;

	if (skewness < 1)
		return 1e100;

	for (i = 0; i < 6; i++)
		sq[i] = coeff[i] * coeff[i];
	for (i = 0; i < 4; i++) {
		d = 2 * coeff[i];
		k = i + 1;
		for (j = i + 2; j < 6; j += 2, k++)
			sq[k] += d * coeff[j];
	}
	s = skewness;
	res = 0.;
	res += s * sq[3] / 35.;
	res += sq[2] / s / 35.;
	s *= (skewness * skewness);
	res += s * sq[4] / 27.;
	res += sq[1] / s / 27.;
	s *= (skewness * skewness);
	res += s * sq[5] / 11.;
	res += sq[0] / s / 11.;
	return res;
}

#ifdef __cplusplus
}
#endif

#endif /* !_STAGE2_H_ */
